import org.apache.commons.httpclient.*
import org.apache.commons.httpclient.methods.*;
import org.apache.commons.httpclient.protocol.*;
import org.apache.commons.httpclient.auth.AuthScope;
import com.urbancode.commons.util.https.OpenSSLProtocolSocketFactory;


final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'
final def out = System.out

final def getStepProperties = {
    def props = [:]

    def ahptoolProc = [ahptool, "getStepProperties"].execute()
    ahptoolProc.consumeProcessErrorStream(System.out);
    def propertiesXml = ahptoolProc.text
    ahptoolProc.waitFor()
    if (ahptoolProc.exitValue()) {
        throw new Exception("Failed to get step properties from ahptool: " + propertiesXml)
    }
    new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
        props[propElem.@name.text()] = propElem.text()
    }

    return props;
}

//-------------------------------------------------------------------------
// Scripts variables
//=========================================================================

final def props = getStepProperties()


System.out.println "Server:\t"+props['integration/rallyUrl']
System.out.println "Defect Pattern:\t"+ props['defectPattern']
System.out.println "Task Pattern:\t"+ props['taskPattern']
System.out.println "User Story Pattern:\t"+ props['usPattern']
            


//post to rally
String base_url = new URL(props['integration/rallyUrl']).getHost()
String url = props['integration/rallyUrl']+"/slm/webservice/"+props['integration/rallyVer']+"/"
String user = props['integration/rallyUser']
String password = props['integration/rallyPass']?props['integration/rallyPass']:props['integration/passScript']
String proxyHost = props['integration/proxyHost']
String proxyPort = props['integration/proxyPort']
String proxyPass = props['integration/proxyPass']
String proxyUser = props['integration/proxyUser']

final def defectArtifactPattern = java.util.regex.Pattern.compile(props['defectPattern'])
final def taskArtifactPattern = java.util.regex.Pattern.compile(props['taskPattern'])
final def usArtifactPattern = java.util.regex.Pattern.compile(props['usPattern'])

HttpClient client = new HttpClient()
client.getParams().setAuthenticationPreemptive(true)
Credentials defaultcreds = new UsernamePasswordCredentials(user, password)
client.getState().setCredentials(new AuthScope(base_url, 443), defaultcreds)

ProtocolSocketFactory socketFactory = new OpenSSLProtocolSocketFactory()
Protocol https = new Protocol("https", socketFactory, 443)
Protocol.registerProtocol("https", https)

// setup proxy if neccessary
if (proxyHost) {
    client.getHostConfiguration().setProxy(proxyHost, Integer.valueOf(proxyPort));
    if (proxyUser && proxyPass) {
        client.getState().setProxyCredentials(AuthScope.ANY,
                new UsernamePasswordCredentials(proxyUser, proxyPass));
    }
}


//---------------------------------------------------------------------------
// Rally Workspace and objid resolution stuff
//===========================================================================
def final getWorkspace = {
    HttpMethod workspaceMethod = null
    try {
        workspaceMethod= new GetMethod(url + "subscription")
        workResult = client.executeMethod(workspaceMethod)
        workResponse = workspaceMethod.getResponseBodyAsString()
        workXml = new XmlParser().parseText(workResponse)
        return workXml.Workspaces.Workspace.findAll{it.'@refObjectName'.equals(props['integration/workspace']) }[0].'@ref'
    }
    finally {
       if(workspaceMethod != null) {
           workspaceMethod.releaseConnection()
       }
    }
}
    



//-----------------------------------------------------------------------------
// Heres where the magic happens
// =============================================================================
workspaceRef = getWorkspace()

try {
    def sourceChangesProc = [ahptool, "getSourceChanges"].execute()
    sourceChangesProc.consumeProcessErrorStream(out);
    def changesXml = sourceChangesProc.text
    sourceChangesProc.waitFor()
    if (sourceChangesProc.exitValue()) {
        throw new Exception("Failed to get build life source changes from ahptool: $changesXml")
    }
    def defectIdSet = [] as Set
    def defectIdToChangeSetListMap = [:]
    def taskIdSet = [] as Set
    def taskIdToChangeSetListMap = [:]
    def usIdSet = [] as Set
    def usIdToChangeSetListMap = [:]
    new XmlSlurper().parseText(changesXml)."change-set".each { changeSetElem ->
    
        def defectMatcher = defectArtifactPattern.matcher(changeSetElem.comment.text());
        def taskMatcher = taskArtifactPattern.matcher(changeSetElem.comment.text());
        def usMatcher = usArtifactPattern.matcher(changeSetElem.comment.text());
        while (defectMatcher.find()) {
            def defectId
            if (defectMatcher.groupCount() > 0 ) {
                // they specified a '(...)' group within the pattern, use that as the bug-id
                defectId = defectMatcher.group(1)
            }
            else {
                // use the whole matching substring as the bug-id
                defectId = defectMatcher.group()
            }
            
            out.println("Found defect identifier in change comment: $defectId")
            defectIdSet.add(defectId)
            def changeSetList = defectIdToChangeSetListMap[defectId]
            if (!changeSetList) {
                changeSetList = []
                defectIdToChangeSetListMap[defectId] = changeSetList
            }
            changeSetList.add(changeSetElem."comment".text())
        }
        while (taskMatcher.find()) {
            def taskId
            if (taskMatcher.groupCount() > 0 ) {
                // they specified a '(...)' group within the pattern, use that as the bug-id
                taskId = taskMatcher.group(1)
            }
            else {
                // use the whole matching substring as the bug-id
                taskId = taskMatcher.group()
            }
            
            out.println("Found task identifier in change comment: $taskId")
            taskIdSet.add(taskId)
            def changeSetList = taskIdToChangeSetListMap[taskId]
            if (!changeSetList) {
                changeSetList = []
                taskIdToChangeSetListMap[taskId] = changeSetList
            }
            changeSetList.add(changeSetElem."comment".text())
        }
        while (usMatcher.find()) {
            def usId
            if (usMatcher.groupCount() > 0 ) {
                // they specified a '(...)' group within the pattern, use that as the bug-id
                usId = usMatcher.group(1)
            }
            else {
                // use the whole matching substring as the bug-id
                usId = usMatcher.group()
            }
            
            out.println("Found User Story identifier in change comment: $usId")
            usIdSet.add(usId)
            def changeSetList = usIdToChangeSetListMap[usId]
            if (!changeSetList) {
                changeSetList = []
                usIdToChangeSetListMap[usId] = changeSetList
            }
            changeSetList.add(changeSetElem."comment".text())
        }
    } 
    if (usIdSet.size() == 0 && taskIdSet.size() == 0 && defectIdSet.size() == 0) {
        out.println("No Rally artifact references found in build life changes.")
    }
    else {
        for (def defectId in defectIdSet) {
            def defectData=null
            HttpMethod method = null
            def response=null
            try {
               method = new GetMethod(url + "defect?workspace=" + URLEncoder.encode(workspaceRef, "UTF-8") + "&query=" + URLEncoder.encode("(FormattedID = " + defectId+ ")" ,"UTF-8") + "&" + URLEncoder.encode("fetch", "UTF-8") + "=" + URLEncoder.encode("true", "UTF-8"))
               result = client.executeMethod(method)
               defectData = method.getResponseBodyAsString()
            }
            finally {
                if(method != null) {
                   method.releaseConnection();
                }
            }
            def defectResult = new XmlParser().parseText(defectData).Results.Object;
            if (defectResult.size() > 0) {
                // Generate XML doc
                def writer = new StringWriter() 
                def builder = new groovy.xml.MarkupBuilder(writer) 
                def changeLogText = defectIdToChangeSetListMap.get(defectId).get(0)
                out.println("Adding change comment for: $defectId")
                builder.ConversationPost() {
                    Artifact(defectResult.ObjectID[0].text())
                    Text(changeLogText + (props['comment']?:''))
                    Workspace(workspaceRef)
                }
                // create and setup httpClient
                method=null
                try {
                    method = new PostMethod(url+"conversationpost/create")
                    method.setRequestEntity((RequestEntity)new StringRequestEntity(writer.toString(), null, null));
                    result = client.executeMethod(method)
                    response = method.getResponseBodyAsString()
                }
                finally {
                    if(method != null) {
                        method.releaseConnection()
                    }
                }
               //check results
               if(response == null) {
                  throw new Exception("HTTP Response not found")
               }
               def updateResult = new XmlParser().parseText(response);
               def errors = updateResult.Errors.OperationResultError.size();
               def warnings = updateResult.Warnings.OperationResultWarning.size();
               if(errors == 0 && warnings == 0 && result >=200 && result <300) {
                  System.out.println("Successfully created comment! ObjectRef =" + updateResult.Object.'@ref');
               }
               else {
                   System.out.println("ERROR!!! Rally Server could not create comment! ")
                   updateResult.Errors.OperationResultError.each { System.out.println(it.text()) }
                   updateResult.Warnings.OperationResultWarning.each { System.out.println(it.text()) }
               }
           }
           else {
              out.println("Unable to find Rally defect $defectId")
           }
        }
        for (def taskId in taskIdSet) {
            def taskData=null
            HttpMethod method = null
            def response=null
            try {
               method = new GetMethod(url + "task?workspace=" + URLEncoder.encode(workspaceRef, "UTF-8") + "&query=" + URLEncoder.encode("(FormattedID = " + taskId+ ")" ,"UTF-8") + "&" + URLEncoder.encode("fetch", "UTF-8") + "=" + URLEncoder.encode("true", "UTF-8"))
               result = client.executeMethod(method)
               taskData = method.getResponseBodyAsString()
            }
            finally {
                if(method != null) {
                   method.releaseConnection();
                }
            }
            def taskResult = new XmlParser().parseText(taskData).Results.Object;
            if (taskResult.size() > 0) {
                // Generate XML doc
                def writer = new StringWriter() 
                def builder = new groovy.xml.MarkupBuilder(writer) 
                def changeLogText = taskIdToChangeSetListMap.get(taskId).get(0)
                out.println("Adding change comment for: $taskId")
                builder.ConversationPost(){
                    Artifact(taskResult.ObjectID[0].text())
                    Text(changeLogText + (props['comment']?:''))
                    Workspace(workspaceRef)
                }
                // create and setup httpClient
                method=null
                try {
                    method = new PostMethod(url+"conversationpost/create")
                    method.setRequestEntity((RequestEntity)new StringRequestEntity(writer.toString(), null, null));
                    result = client.executeMethod(method)
                    response = method.getResponseBodyAsString()
                }
                finally {
                    if(method != null) {
                        method.releaseConnection()
                    }
                }
               //check results
               if(response == null) {
                  throw new Exception("HTTP Response not found")
               }
               def updateResult = new XmlParser().parseText(response);
               def errors = updateResult.Errors.OperationResultError.size();
               def warnings = updateResult.Warnings.OperationResultWarning.size();
               if(errors == 0 && warnings == 0 && result >=200 && result <300) {
                   System.out.println("Successfully created comment! ObjectRef =" + updateResult.Object.'@ref');
               }
               else {
                   System.out.println("ERROR!!! Rally Server could not create comment! ")
                   updateResult.Errors.OperationResultError.each { System.out.println(it.text()) }
                   updateResult.Warnings.OperationResultWarning.each { System.out.println(it.text()) }
                }
            }
            else {
                out.println("Unable to find Rally task $taskId")
            }
        }
        for (def usId in usIdSet) {
            def usData=null
            HttpMethod method = null
            def response=null
            try {
               method = new GetMethod(url + "hierarchicalrequirement?workspace=" + URLEncoder.encode(workspaceRef, "UTF-8") + "&query=" + URLEncoder.encode("(FormattedID = " + usId+ ")" ,"UTF-8") + "&" + URLEncoder.encode("fetch", "UTF-8") + "=" + URLEncoder.encode("true", "UTF-8"))
               result = client.executeMethod(method)
               usData = method.getResponseBodyAsString()
            }
            finally {
                if(method != null) {
                   method.releaseConnection();
                }
            }
            def usResult = new XmlParser().parseText(usData).Results.Object;
            if (usResult.size() > 0) {
                // Generate XML doc
                def writer = new StringWriter() 
                def builder = new groovy.xml.MarkupBuilder(writer) 
                def changeLogText = usIdToChangeSetListMap.get(usId).get(0)
                out.println("Adding change comment for: $usId")
                builder.ConversationPost(){
                    Artifact(usResult.ObjectID[0].text())
                    Text(changeLogText + (props['comment']?:''))
                    Workspace(workspaceRef)
                }
                // create and setup httpClient
                method=null
                try {
                    method = new PostMethod(url+"conversationpost/create")
                    method.setRequestEntity((RequestEntity)new StringRequestEntity(writer.toString(), null, null));
                    result = client.executeMethod(method)
                    response = method.getResponseBodyAsString()
                }
                finally {
                    if(method != null) {
                        method.releaseConnection()
                    }
                }
               //check results
               if(response == null) {
                  throw new Exception("HTTP Response not found")
               }
               def updateResult = new XmlParser().parseText(response);
               def errors = updateResult.Errors.OperationResultError.size();
               def warnings = updateResult.Warnings.OperationResultWarning.size();
               if(errors == 0 && warnings == 0 && result >=200 && result <300) {
                   System.out.println("Successfully created comment! ObjectRef =" + updateResult.Object.'@ref');
               }
               else {
                   System.out.println("ERROR!!! Rally Server could not create comment! ")
                   updateResult.Errors.OperationResultError.each { System.out.println(it.text()) }
                   updateResult.Warnings.OperationResultWarning.each { System.out.println(it.text()) }
               }
            }
            else {
                out.println("Unable to find Rally User Story $usId")
            }
        }
    } 
}
catch (Exception e) {
    e.printStackTrace()
    System.exit(1)
}
