import org.apache.commons.httpclient.*
import org.apache.commons.httpclient.methods.*;
import org.apache.commons.httpclient.protocol.*;
import org.apache.commons.httpclient.auth.AuthScope;
import org.jsoup.Jsoup;
import com.urbancode.commons.util.https.OpenSSLProtocolSocketFactory;



final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'
final def out = System.out

final def getStepProperties = {
    def props = [:]

    def ahptoolProc = [ahptool, "getStepProperties"].execute()
    ahptoolProc.consumeProcessErrorStream(System.out);
    def propertiesXml = ahptoolProc.text
    ahptoolProc.waitFor()
    if (ahptoolProc.exitValue()) {
        throw new Exception("Failed to get step properties from ahptool: " + propertiesXml)
    }
    new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
        props[propElem.@name.text()] = propElem.text()
    }

    return props;
}

//-------------------------------------------------------------------------
// Scripts variables
//=========================================================================

final def props = getStepProperties()


System.out.println "Server:\t"+props['integration/rallyUrl']
System.out.println "Defect Pattern:\t"+ props['defectPattern']
System.out.println "Task Pattern:\t"+ props['taskPattern']
System.out.println "User Story Pattern:\t"+ props['usPattern']
            

final def defectArtifactPattern = java.util.regex.Pattern.compile(props['defectPattern'])
final def taskArtifactPattern = java.util.regex.Pattern.compile(props['taskPattern'])
final def usArtifactPattern = java.util.regex.Pattern.compile(props['usPattern'])

final def workDir = new File('.').canonicalFile


String base_url = new URL(props['integration/rallyUrl']).getHost()
String url = props['integration/rallyUrl']+"/slm/webservice/"+props['integration/rallyVer']+ "/"
String user = props['integration/rallyUser']
String password = props['integration/rallyPass']?props['integration/rallyPass']:props['integration/passScript']
String proxyHost = props['integration/proxyHost']
String proxyPort = props['integration/proxyPort']
String proxyPass = props['integration/proxyPass']
String proxyUser = props['integration/proxyUser']

HttpClient client = new HttpClient()
client.getParams().setAuthenticationPreemptive(true)
Credentials defaultcreds = new UsernamePasswordCredentials(user, password)
client.getState().setCredentials(new AuthScope(base_url, 443), defaultcreds)

ProtocolSocketFactory socketFactory = new OpenSSLProtocolSocketFactory()
Protocol https = new Protocol("https", socketFactory, 443)
Protocol.registerProtocol("https", https)

// setup proxy if neccessary
if (proxyHost) {
	client.getHostConfiguration().setProxy(proxyHost, Integer.valueOf(proxyPort));
	if (proxyUser && proxyPass) {
		client.getState().setProxyCredentials(AuthScope.ANY,
				new UsernamePasswordCredentials(proxyUser, proxyPass));
	}
}



//---------------------------------------------------------------------------
// Rally Workspace and objid resolution stuff
//===========================================================================
def final getWorkspace = {
    HttpMethod workspaceMethod=null
    try {
        workspaceMethod= new GetMethod(url + "subscription")
        workResult = client.executeMethod(workspaceMethod)
        workResponse = workspaceMethod.getResponseBodyAsString()
        workXml = new XmlParser().parseText(workResponse)
        return workXml.Workspaces.Workspace.findAll{it.'@refObjectName'.equals(props['integration/workspace']) }[0].'@ref'
    }
    finally {
       if(workspaceMethod != null) {
           workspaceMethod.releaseConnection()
       }
    }
}
    


//-----------------------------------------------------------------------
// Script Content
//========================================================================
out.println "Getting Workspace"
workspaceRef = getWorkspace()
out.println "Workspace Ref = ${workspaceRef}"
try {
    def sourceChangesProc = [ahptool, "getSourceChanges"].execute()
    sourceChangesProc.consumeProcessErrorStream(out);
    def changesXml = sourceChangesProc.text
    sourceChangesProc.waitFor()
    if (sourceChangesProc.exitValue()) {
        throw new Exception("Failed to get build life source changes from ahptool: $changesXml")
    }
    out.println "Parsing Changelog"
    def defectIdSet = [] as Set
    def defectIdToChangeSetListMap = [:]
    def taskIdSet = [] as Set
    def taskIdToChangeSetListMap = [:]
    def usIdSet = [] as Set
    def usIdToChangeSetListMap = [:]
    new XmlSlurper().parseText(changesXml)."change-set".each { changeSetElem ->
    
        def defectMatcher = defectArtifactPattern.matcher(changeSetElem.comment.text());
        def taskMatcher = taskArtifactPattern.matcher(changeSetElem.comment.text());
        def usMatcher = usArtifactPattern.matcher(changeSetElem.comment.text());
        while (defectMatcher.find()) {
            def defectId
            if (defectMatcher.groupCount() > 0 ) {
                // they specified a '(...)' group within the pattern, use that as the bug-id
                defectId = defectMatcher.group(1)
            }
            else {
                // use the whole matching substring as the bug-id
                defectId = defectMatcher.group()
            }
            
            out.println("Found defect identifier in change comment: $defectId")
            defectIdSet.add(defectId)
            def changeSetList = defectIdToChangeSetListMap[defectId]
            if (!changeSetList) {
                changeSetList = []
                defectIdToChangeSetListMap[defectId] = changeSetList
            }
            changeSetList.add(changeSetElem."anthill-id".text())
        }
        while (taskMatcher.find()) {
            def taskId
            if (taskMatcher.groupCount() > 0 ) {
                // they specified a '(...)' group within the pattern, use that as the bug-id
                taskId = taskMatcher.group(1)
            }
            else {
                // use the whole matching substring as the bug-id
                taskId = taskMatcher.group()
            }
            
            out.println("Found task identifier in change comment: $taskId")
            taskIdSet.add(taskId)
            def changeSetList = taskIdToChangeSetListMap[taskId]
            if (!changeSetList) {
                changeSetList = []
                taskIdToChangeSetListMap[taskId] = changeSetList
            }
            changeSetList.add(changeSetElem."anthill-id".text())
        }
        while (usMatcher.find()) {
            def usId
            if (usMatcher.groupCount() > 0 ) {
                // they specified a '(...)' group within the pattern, use that as the bug-id
                usId = usMatcher.group(1)
            }
            else {
                // use the whole matching substring as the bug-id
                usId = usMatcher.group()
            }
            
            out.println("Found User Story identifier in change comment: $usId")
            usIdSet.add(usId)
            def changeSetList = usIdToChangeSetListMap[usId]
            if (!changeSetList) {
                changeSetList = []
                usIdToChangeSetListMap[usId] = changeSetList
            }
            changeSetList.add(changeSetElem."anthill-id".text())
        }
    } 
    if (usIdSet.size() == 0 && taskIdSet.size() == 0 && defectIdSet.size() == 0) {
        out.println("No Rally artifact references found in build life changes.")
    }
    else {
        out.println("Creating Issues.xml")
        def issuesXml = new java.io.StringWriter()
        new groovy.xml.MarkupBuilder(issuesXml).issues() {
            for (def defectId in defectIdSet) {
                out.println("Adding ${defectId} to Issues.xml")
                def defectData=null
                HttpMethod method = null
                try {
                    method = new GetMethod(url + "defect?workspace=" + URLEncoder.encode(workspaceRef, "UTF-8") + "&query=" + URLEncoder.encode("(FormattedID = " + defectId+ ")" ,"UTF-8") + "&" + URLEncoder.encode("fetch", "UTF-8") + "=" + URLEncoder.encode("true", "UTF-8"))
                    result = client.executeMethod(method)
                    defectData = method.getResponseBodyAsString()
                }
                finally {
                    if(method != null) {
                        method.releaseConnection();
                    }
                }
                def defectResult = new XmlParser().parseText(defectData).Results.Object;
                if (defectResult.size() > 0) {
                    out.println("Found Rally defect $defectId - ${defectResult.'@refObjectName'}")
                    
                    // create the issue element
                    def changeId = defectIdToChangeSetListMap.get(defectId).get(0)
                    issue(id:defectId, "issue-tracker":"Rally", "change-id":changeId) {
                        // create the name element
                        name(defectResult.'@refObjectName'[0])
                        // create the type element
                        // type would be the tracker name in TeamForge, but their SOAP services do not allow you to get
                        // the tracker from the defect id
                        type(defectResult.'@type'[0])
                        // create the description element
                        description(Jsoup.parse(defectResult.'Description'.text()).text())
                        // create the status element
                        status(defectResult.State.text())
                    }
                }
                else {
                    out.println("Unable to find Rally defect $defectId")
                }
            }
            for (def taskId in taskIdSet) {
                out.println("Adding ${taskId} to Issues.xml")
                def taskData=null
                HttpMethod method = null
                try {
                    method = new GetMethod(url + "task?workspace=" + URLEncoder.encode(workspaceRef, "UTF-8") + "&query=" + URLEncoder.encode("(FormattedID = " + taskId+ ")" ,"UTF-8") + "&" + URLEncoder.encode("fetch", "UTF-8") + "=" + URLEncoder.encode("true", "UTF-8"))
                    result = client.executeMethod(method)
                    taskData = method.getResponseBodyAsString()
                }
                finally {
                    if(method != null) {
                        method.releaseConnection();
                    }
                }
                def taskResult = new XmlParser().parseText(taskData).Results.Object;
                if (taskResult.size() > 0) {
                    out.println("Found Rally task $taskId - ${taskResult.'@refObjectName'}")
                    
                    // create the issue element
                    def changeId = taskIdToChangeSetListMap.get(taskId).get(0)
                    issue(id:taskId, "issue-tracker":"Rally", "change-id":changeId) {
                        // create the name element
                        name(taskResult.'@refObjectName'[0])
                        // create the type element
                        // type would be the tracker name in TeamForge, but their SOAP services do not allow you to get
                        // the tracker from the task id
                        type(taskResult.'@type'[0])
                        // create the description element
                        description(Jsoup.parse(taskResult.'Description'.text()).text())
                        // create the status element
                        status(taskResult.State.text())
                    }
                }
                else {
                    out.println("Unable to find Rally task $taskId")
                }
            }
            for (def usId in usIdSet) {
                def usData=null
                out.println("Adding ${usId} to Issues.xml")
                HttpMethod method = null
                try {
                    method = new GetMethod(url + "hierarchicalrequirement?workspace=" + URLEncoder.encode(workspaceRef, "UTF-8") + "&query=" + URLEncoder.encode("(FormattedID = " + usId+ ")" ,"UTF-8") + "&" + URLEncoder.encode("fetch", "UTF-8") + "=" + URLEncoder.encode("true", "UTF-8"))
                    result = client.executeMethod(method)
                    usData = method.getResponseBodyAsString()
                }
                finally {
                    if(method != null) {
                        method.releaseConnection();
                    }
                }
                def usResult = new XmlParser().parseText(usData).Results.Object;
                if (usResult.size() > 0) {
                    out.println("Found Rally User Story $usId - ${usResult.'@refObjectName'}")
                    
                    // create the issue element
                    def changeId = usIdToChangeSetListMap.get(usId).get(0)
                    issue(id:usId, "issue-tracker":"Rally", "change-id":changeId) {
                        // create the name element
                        name(usResult.'@refObjectName'[0])
                        // create the type element
                        // type would be the tracker name in TeamForge, but their SOAP services do not allow you to get
                        // the tracker from the us id
                        type('User Story')
                        // create the description element
                        description(Jsoup.parse(usResult.'Description'.text()).text())
                    }
                }
                else {
                    out.println("Unable to find Rally User Story $usId")
                }
            }
        } 
        out.println("uploading Issues.xml")
       
        def addIssuesProc = new ProcessBuilder([ahptool, 'addIssues', '-']).directory(workDir).start()

        addIssuesProc.consumeProcessOutput(out, out);
        addIssuesProc.withWriter{ it << issuesXml }

        addIssuesProc.waitFor()

        if (addIssuesProc.exitValue()) {
            println("Rally results upload to server failed with exit code: ${addIssuesProc.exitValue()}")
            System.exit(addIssuesProc.exitValue())
        }
        else {
            println("Rally results uploaded successfully")
        }
    }
}
catch (Exception e) {
    e.printStackTrace()
    System.exit(1)
}